# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- base node
# :Created:   mer 07 feb 2018 16:17:02 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

import inspect

from ..processor.util import Line, Part


class TargetNode:
    """This is the common ancestor of all the JS AST nodes."""

    """The associated Python AST node"""
    py_node = None

    """The Transformer instance which is managing this one"""
    transformer = None

    """The final arguments passed to the emit method defined on the
    subclasses"""
    transformed_args = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.options = kwargs

    def __str__(self):
        return ''.join(str(x) for x in self.serialize())

    def _chain(self, items):
        for i in self._expand(items):
            if inspect.isgenerator(i):
                yield from i
            else:
                yield i

    def _expand(self, items):
        for i in items:
            if isinstance(i, TargetNode):
                yield from i.serialize()
            else:
                yield i

    def emit(self):
        """This is the main output definition method. Is reimplemented by the
        subclasses."""

    @classmethod
    def final(cls, *transformed_args, **options):
        tn = cls(**options)
        tn.transformed_args = transformed_args
        return tn

    def line(self, item, indent=False, delim=False, name=None):
        if isinstance(item, Line):
            item.indent += int(indent)
            l = item
        elif isinstance(item, (tuple, list)):
            item = tuple(self._chain(item))
            l = Line(self, item, indent, delim, name)
        else:
            l = Line(self, item, indent, delim, name)
        return l

    def lines(self, items, *, indent=False, delim=False, name=None):
        if not isinstance(items, (tuple, list)):
            items = (items,)
        for i in self._chain(items):
            yield self.line(i, indent=indent, delim=delim, name=name)

    def part(self, *items, name=None):
        it = tuple(self._expand(items))
        if len(it) == 1 and isinstance(it[0], Line):
            result = it[0].item
        else:
            result = Part(self, *it, name=name)
        return result

    def serialize(self):
        for a in self.emit(*self.transformed_args, **self.options):
            yield from a.serialize()


class JSNode(TargetNode):
    pass


class JSStatement(JSNode):
    pass



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- noops
# :Created:   gio 08 feb 2018 00:50:00 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .base import JSNode
from ..processor.util import delimited_multi_line


class JSPass(JSNode):
    def emit(self):
        return []


class JSCommentBlock(JSNode):
    def emit(self, text):
        assert text.find('*/') == -1
        yield from self.lines(
            delimited_multi_line(self, text, '/*', '*/', True))



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- compatibility
# :Created:   lun 30 mar 2020, 01:48:33
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2020 Alberto Berti
#

import ast
import sys

is_py36 = sys.version_info >= (3, 6)

is_py39 = sys.version_info >= (3, 9)

if is_py36:
    assign_types = (ast.Assign, ast.AnnAssign)
else:
    assign_types = (ast.Assign,)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- Aggregated statements
# :Created:   gio 08 feb 2018 02:53:30 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

import itertools

from .base import JSNode
from .noops import JSCommentBlock
from .statements import JSImport, JSVarStatement


class JSStatements(JSNode):

    def __iadd__(self, other):
        self.transformed_args.extend(other.transformed_args)
        return self

    def emit(self, statements):
        for s in statements:
            yield s

    def squash(self, args):
        for a in args:
            if isinstance(a, JSStatements):
                yield from a.transformed_args
            else:
                yield a

    def reordered_args(self, args):
        """Reorder the args to keep the imports and vars always at the top."""
        args = list(self.squash(args))
        imports = []
        vars_ = []
        others = []
        for a in args:
            if isinstance(a, JSImport):
                imports.append(a)
            elif isinstance(a, JSVarStatement) and \
                 not a.options.get('unmovable', False):
                vars_.append(a)
            else:
                others.append(a)

        others_first = []
        others_after = []
        # if the others start with some comments, put those at the top
        start_trigger = False
        for s in others:
            if isinstance(s, JSCommentBlock) and not start_trigger:
                others_first.append(s)
            else:
                others_after.append(s)
                start_trigger = True

        return itertools.chain(others_first, imports, vars_, others_after)

    def serialize(self):
        for a in self.emit(self.reordered_args(self.transformed_args)):
            yield from self.lines(a.serialize(), delim=True)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- js ast
# :Created:   mer 07 feb 2018 16:14:21 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .base import TargetNode
from .blocks import (
    JSForIterableStatement,
    JSForStatement,
    JSForeachStatement,
    JSForofStatement,
    JSIfStatement,
    JSTryCatchFinallyStatement,
    JSWhileStatement,
)
from .functions import (
    JSArrowFunction,
    JSAsyncFunction,
    JSFunction,
    JSGenFunction,
)

from .literals import (
    JSDict,
    JSFalse,
    JSList,
    JSLiteral,
    JSNull,
    JSNum,
    JSStr,
    JSTrue,
)

from .noops import (
    JSCommentBlock,
    JSPass,
)

from .operators import (
    JSIs,
    JSLeftSideUnaryOp,
    JSOpAdd,
    JSOpAnd,
    JSOpBitAnd,
    JSOpBitOr,
    JSOpBitXor,
    JSOpDiv,
    JSOpGt,
    JSOpGtE,
    JSOpIn,
    JSOpInstanceof,
    JSOpInvert,
    JSOpLShift,
    JSOpLt,
    JSOpLtE,
    JSOpMod,
    JSOpMult,
    JSOpNot,
    JSOpOr,
    JSOpRShift,
    JSOpStrongEq,
    JSOpStrongNotEq,
    JSOpSub,
    JSOpTypeof,
    JSOpUSub,
    JSRest,
)

from .statements import (
    JSAugAssignStatement,
    JSAwait,
    JSBreakStatement,
    JSContinueStatement,
    JSDefaultImport,
    JSDeleteStatement,
    JSDependImport,
    JSExport,
    JSExportDefault,
    JSExpressionStatement,
    JSImport,
    JSLetStatement,
    JSNamedImport,
    JSReturnStatement,
    JSStarImport,
    JSThrowStatement,
    JSVarStatement,
    JSYield,
    JSYieldStar,
)

from .expressions import (
    JSAssignmentExpression,
    JSAttribute,
    JSBinOp,
    JSCall,
    JSExpression,
    JSIfExp,
    JSKeySubscript,
    JSMultipleArgsOp,
    JSName,
    JSNewCall,
    JSSubscript,
    JSSuper,
    JSTaggedTemplate,
    JSTemplateLiteral,
    JSThis,
    JSUnaryOp,
)

from .classes import (
    JSAsyncMethod,
    JSClass,
    JSClassConstructor,
    JSClassMember,
    JSGenMethod,
    JSGetter,
    JSMethod,
    JSSetter,
)

from .bodies import (
    JSStatements,
)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- statements
# :Created:   gio 08 feb 2018 01:47:43 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .base import JSStatement
from .util import _check_keywords
from ..processor.util import delimited


class JSVarDeclarer(JSStatement):

    def with_kind(self, kind, keys, values):
        for key in keys:
            _check_keywords(self, key)
        assert len(keys) > 0
        assert len(keys) == len(values)

        arr = ['%s ' % kind]
        for i in range(len(keys)):
            if i > 0:
                arr.append(', ')
            arr.append(keys[i])
            if values[i] is not None:
                arr.append(' = ')
                arr.append(values[i])
        yield self.part(*arr)


class JSVarStatement(JSVarDeclarer):
    def emit(self, keys, values, unmovable=False):
        yield from self.with_kind('var', keys, values)


class JSLetStatement(JSVarDeclarer):
    def emit(self, keys, values, unmovable=True):
        yield from self.with_kind('let', keys, values)


class JSAugAssignStatement(JSStatement):
    def emit(self, target, op, value):
        yield self.part(target, ' ', op, '= ', value, name=str(target))


class JSReturnStatement(JSStatement):
    def emit(self, value):
        if value:
            result = self.line(['return ', value], delim=True)
        else:
            result = self.line('return', delim=True)
        yield result


class JSBreakStatement(JSStatement):
    def emit(self):
        yield self.part('break')


class JSContinueStatement(JSStatement):
    def emit(self):
        yield self.part('continue')


class JSDeleteStatement(JSStatement):
    def emit(self, value):
        yield self.line(['delete ', value], delim=True)


class JSThrowStatement(JSStatement):
    def emit(self, obj):
        yield self.line(['throw ', obj], delim=True)


class JSYield(JSStatement):

    def emit(self, expr):
        yield self.part('yield ', expr)


class JSYieldStar(JSStatement):

    def emit(self, expr):
        yield self.part('yield* ', expr)


class JSAwait(JSStatement):

    def emit(self, value):
        yield self.part('await ', value)


class JSImport(JSStatement):
    pass


class JSDependImport(JSImport):
    def emit(self, module):
        yield self.line(['System.import(', "'", module, "'", ')'], delim=True)


class JSNamedImport(JSImport):
    def emit(self, module, names):
        js_names = []
        for name, alias in sorted(names):
            if alias:
                js_names.append(self.part(name, ' as ', alias))
            else:
                js_names.append(self.part(name))

        yield self.line(['import {', *delimited(', ', js_names),
                         "} from '", module, "'"], delim=True)


class JSStarImport(JSImport):
    def emit(self, module, name):
        yield self.line(['import * as ', name, " from '", module, "'"],
                        delim=True)


class JSDefaultImport(JSImport):
    def emit(self, module, alias):
        yield self.line(['import ', alias, " from '", module, "'"], delim=True)


class JSExport(JSStatement):
    def emit(self, names):
        yield self.line(['export ', '{', *delimited(', ', names), '}'],
                        delim=True)


class JSExportDefault(JSExport):
    def emit(self, name):
        yield self.line(['export default ', name], delim=True)


class JSExpressionStatement(JSStatement):
    def emit(self, value):
        yield self.part(value)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- literals
# :Created:   mer 07 feb 2018 17:50:03 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

import json

from .base import JSNode
from ..processor.util import delimited, delimited_multi_line


class JSLiteral(JSNode):
    def emit(self, text):
        yield from self.lines(delimited_multi_line(self, text, '', '', False))


class JSDict(JSLiteral):
    def emit(self, keys, values):
        arr = ['{']
        for i in range(len(keys)):
            if i > 0:
                arr.append(', ')
            arr.append(keys[i])
            arr.append(': ')
            arr.append(values[i])
        arr.append('}')
        yield self.part(*arr)


class JSList(JSLiteral):
    def emit(self, elts):
        arr = ['[']
        delimited(', ', elts, dest=arr)
        arr.append(']')
        yield self.part(*arr)


class JSFalse(JSLiteral):
    def emit(self):
        yield self.part('false')


class JSNull(JSLiteral):
    def emit(self):
        yield self.part('null')


class JSNum(JSLiteral):
    def emit(self, x):
        yield self.part(str(x))


class JSStr(JSLiteral):
    def emit(self, s):
        yield self.part(json.dumps(s))


class JSTrue(JSLiteral):
    def emit(self):
        yield self.part('true')



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- block statements
# :Created:   gio 08 feb 2018 02:25:23 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .base import JSStatement


class JSBlock(JSStatement):
    pass


class JSIfStatement(JSBlock):
    def emit(self, test, body, orelse):
        yield self.line(['if (', test, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        if orelse:
            yield self.line(['} else {'])
            yield from self.lines(orelse, indent=True, delim=True)
            yield self.line('}')
        else:
            yield self.line('}')


class JSWhileStatement(JSBlock):
    def emit(self, test, body):
        yield self.line(['while (', test, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForStatement(JSBlock):
    def emit(self, left, test, right, body):
        yield self.line(['for (', left, '; ', test, '; ', right, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForIterableStatement(JSBlock):

    operator = ' of '

    def emit(self, target, source, body):
        yield self.line(['for (var ', self.part(target), self.operator,
                         source, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForeachStatement(JSForIterableStatement):

    operator = ' in '


class JSForofStatement(JSForIterableStatement):
    pass


class JSTryCatchFinallyStatement(JSBlock):
    def emit(self, try_body, target, catch_body, finally_body):
        assert catch_body or finally_body
        yield self.line('try {')
        yield from self.lines(try_body, indent=True, delim=True)
        if catch_body:
            yield self.line(['} catch(', target, ') {'])
            yield from self.lines(catch_body, indent=True, delim=True)
        if finally_body:
            yield self.line(['} finally {'])
            yield from self.lines(finally_body, indent=True, delim=True)
        yield self.line('}')



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- classes
# :Created:   gio 08 feb 2018 02:32:04 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .blocks import JSBlock
from .functions import JSFunction


class JSClass(JSBlock):

    def emit(self, name, super_, methods):
        line = ['class ', name]
        if super_ is not None:
            line += [' extends ', super_]
        line += [' {']
        yield self.line(line)
        yield from self.lines(methods, indent=True, delim=True)
        yield self.line('}')


class JSClassMember(JSFunction):

    def with_kind(self, kind, args, body, acc=None, kwargs=None, static=False):
        if static:
            line = ['static ', kind]
        else:
            line = [kind]
        line += self.fargs(args, acc, kwargs)
        line += ['{']
        yield self.line(line)
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSClassConstructor(JSClassMember):

    def emit(self, args, body, acc=None, kwargs=None):
        yield from self.with_kind('constructor', args, body, acc, kwargs)


class JSMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind(name, args, body, acc, kwargs, static)


class JSAsyncMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind('async ' + name, args, body, acc, kwargs,
                                  static)


class JSGenMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind('* ' + name, args, body, acc, kwargs,
                                  static)


class JSGetter(JSClassMember):

    def emit(self, name, body, static=False):
        yield from self.with_kind('get ' + name, [], body, static=static)


class JSSetter(JSClassMember):

    def emit(self, name, arg, body, static=False):
        yield from self.with_kind('set ' + name, [arg], body, static=static)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- operators
# :Created:   mer 07 feb 2018 17:37:24 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .base import JSNode


class JSOperator(JSNode):
    pass


class JSLeftSideUnaryOp(JSOperator):
    pass


class JSOpIn(JSOperator):
    def emit(self):
        yield self.part('in')


class JSOpAnd(JSOperator):
    def emit(self):
        yield self.part('&&')


class JSOpOr(JSOperator):
    def emit(self):
        yield self.part('||')


class JSOpNot(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('!')


class JSOpInstanceof(JSOperator):
    def emit(self):
        yield self.part('instanceof')


class JSOpTypeof(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('typeof')


class JSOpAdd(JSOperator):
    def emit(self):
        yield self.part('+')


class JSOpSub(JSOperator):
    def emit(self):
        yield self.part('-')


class JSOpMult(JSOperator):
    def emit(self):
        yield self.part('*')


class JSOpDiv(JSOperator):
    def emit(self):
        yield self.part('/')


class JSOpMod(JSOperator):
    def emit(self):
        yield self.part('%')


class JSOpRShift(JSOperator):
    def emit(self):
        yield self.part('>>')


class JSOpLShift(JSOperator):
    def emit(self):
        yield self.part('<<')


class JSOpBitXor(JSOperator):
    def emit(self):
        yield self.part('^')


class JSOpBitAnd(JSOperator):
    def emit(self):
        yield self.part('&')


class JSOpBitOr(JSOperator):
    def emit(self):
        yield self.part('|')


class JSOpInvert(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('~')


class JSOpUSub(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('-')


class JSOpStrongEq(JSOperator):
    def emit(self):
        yield self.part('===')


class JSOpStrongNotEq(JSOperator):
    def emit(self):
        yield self.part('!==')


class JSOpLt(JSOperator):
    def emit(self):
        yield self.part('<')


class JSOpLtE(JSOperator):
    def emit(self):
        yield self.part('<=')


class JSOpGt(JSOperator):
    def emit(self):
        yield self.part('>')


class JSOpGtE(JSOperator):
    def emit(self):
        yield self.part('>=')


JSIs = JSOpStrongEq


class JSRest(JSOperator):
    def emit(self, value):
        yield self.part('...', value)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- expressions
# :Created:   gio 08 feb 2018 02:46:38 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

import re

from .base import JSNode
from .operators import JSLeftSideUnaryOp
from ..processor.util import delimited, delimited_multi_line
from . util import _check_keywords


class JSExpression(JSNode):

    def emit(self, expr):
        yield self.part('(', expr, ')')


class JSAssignmentExpression(JSNode):
    def emit(self, left, right):
        yield self.part(left, ' = ', right)


class JSIfExp(JSNode):
    def emit(self, test, body, orelse):
        yield self.part('(', test, ' ? ', body, ' : ', orelse, ')')


class JSCall(JSNode):

    operator = ''

    def emit(self, func, args, kwargs=None, operator=None):
        operator = operator or self.operator
        kwargs = kwargs or []
        arr = [operator, func, '(']
        fargs = args.copy()
        if kwargs:
            fargs.append(kwargs)
        delimited(', ', fargs, dest=arr)
        arr.append(')')
        yield self.part(*arr)


class JSNewCall(JSCall):

    operator = 'new '


class JSAttribute(JSNode):
    def emit(self, obj, s):
        assert re.search(r'^[a-zA-Z$_][a-zA-Z$_0-9]*$', s)
        _check_keywords(self, s)
        yield self.part(obj, '.', s, name=True)


class JSSubscript(JSNode):
    def emit(self, obj, key):
        yield self.part(self.part(obj, name=True), '[',
                        self.part(key, name=True), ']')


class JSKeySubscript(JSNode):
    def emit(self, key):
        yield self.part('[', self.part(key), ']')


class JSBinOp(JSNode):
    def emit(self, left, op, right):
        yield self.part('(', left, ' ', op, ' ', right, ')')


class JSMultipleArgsOp(JSNode):
    def emit(self, binop, conj, *args):
        assert len(args) > 1
        parts = []
        for ix, arg in enumerate(args):
            if isinstance(binop, (tuple, list)):
                op = binop[ix]
            else:
                op = binop
            if ix > 0:
                parts += [' ', conj, ' ']
            parts += ['(', arg[0], ' ', op, ' ', arg[1], ')']
        yield self.part('(', *parts, ')')


class JSUnaryOp(JSNode):
    def emit(self, op, right):
        assert isinstance(op, JSLeftSideUnaryOp)
        yield self.part('(', op, ' ', right, ')')


class JSName(JSNode):
    def emit(self, name):
        _check_keywords(self, name)
        yield self.part(name, name=True)


class JSTaggedTemplate(JSNode):

    def emit(self, value, func):
        text = list(delimited_multi_line(self, value, '`'))
        func = list(func.serialize())
        yield self.part(*func, *text)


class JSTemplateLiteral(JSNode):

    def emit(self, value):
        yield from delimited_multi_line(self, value, '`')


class JSSuper(JSNode):
    def emit(self):
        yield self.part('super')


class JSThis(JSNode):
    def emit(self):
        yield self.part('this')



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- utils
# :Created:   gio 08 feb 2018 01:27:59 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

JS_KEYWORDS = set([
    'break', 'case', 'catch', 'continue', 'default', 'delete', 'do', 'else',
    'finally', 'for', 'function', 'if', 'in', 'instanceof', 'new', 'return',
    'switch', 'this', 'throw', 'try', 'typeof', 'var', 'void', 'while', 'with',

    'abstract', 'boolean', 'byte', 'char', 'class', 'const',
    'double', 'enum', 'export', 'extends', 'final', 'float', 'goto',
    'implements', 'import', 'int', 'interface', 'long', 'native', 'package',
    'private', 'protected', 'public', 'short', 'static', 'super',
    'synchronized', 'throws', 'transient', 'volatile'])

JS_KEYWORDS_ES6 = JS_KEYWORDS - set(['delete'])


def _check_keywords(target_node, name):
    trans = target_node.transformer
    if trans is not None:
        trans.unsupported(
            target_node.py_node,
            (name in JS_KEYWORDS_ES6 if trans.enable_es6 else name
             in JS_KEYWORDS),
            "Name '%s' is reserved in JavaScript." % name)
    else:
        if name in JS_KEYWORDS:
            raise ValueError("Name %s is reserved in JavaScript." % name)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- functions
# :Created:   gio 08 feb 2018 02:29:14 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Alberto Berti
#

from .blocks import JSBlock
from ..processor.util import delimited


class JSFunction(JSBlock):

    begin = 'function '
    bet_args_n_body = ''

    def fargs(self, args, acc=None, kwargs=None):
        result = []
        result.append('(')
        js_args = args.copy()
        if kwargs:
            js_args.append(self.part('{', *delimited(', ', kwargs), '}={}'))
        if acc:
            js_args.append(acc)
        delimited(', ', js_args, dest=result)
        result.append(') ')
        return result

    def emit(self, name, args, body, acc=None, kwargs=None):
        line = [self.begin]
        if name is not None:
            line.append(name)
        line += self.fargs(args, acc, kwargs)
        line += self.bet_args_n_body
        line += ['{']
        yield self.line(line, name=str(name))
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSAsyncFunction(JSFunction):

    begin = 'async function '


class JSGenFunction(JSFunction):

    begin = 'function* '


class JSArrowFunction(JSFunction):

    begin = ''
    bet_args_n_body = '=> '

    def emit(self, name, args, body, acc=None, kwargs=None):
        if name:
            # TODO: split this into an assignment + arrow function
            line = [name, ' = ']
        else:
            line = []
        line += self.fargs(args, acc, kwargs)
        line += self.bet_args_n_body
        line += ['{']
        yield self.line(line)
        yield from self.lines(body, indent=True, delim=True)
        if name:
            yield self.line('}', delim=True)
        else:
            yield self.part('}')
